local function orb_msg(world, orb)
  world:show_message("You found the " .. orb .. " elemental orb!\n" ..
    "Take it to your lab to transmute it, brew a potion and gain new abilities by drinking it!"
  )
end

function event_collectEarthOrb(world, player, orb)
  G.sounds.orb:play()
  world.world:killEntity(orb.entity)
  table.insert(world.inventory.orbs, {
    name = "earth",
    sprite = G.animRepo:newSprite("EarthOrb_Idle")
  })
  orb_msg(world, "Earth")
end

function event_collectEtherOrb(world, player, orb)
  G.sounds.orb:play()
  world.world:killEntity(orb.entity)
  table.insert(world.inventory.orbs, {
    name = "ether",
    sprite = G.animRepo:newSprite("EtherOrb_Idle")
  })
  world:show_message("You found the Ether elemental orb!\n" ..
    "This is the final element you need for the philosopher's stone. Take it home, and get to work. Quick!"
  )
end

function event_collectFireOrb(world, player, orb)
  G.sounds.orb:play()
  world.world:killEntity(orb.entity)
  table.insert(world.inventory.orbs, {
    name = "fire",
    sprite = G.animRepo:newSprite("FireOrb_Idle")
  })
  orb_msg(world, "Fire")
end

function event_collectWaterOrb(world, player, orb)
  G.sounds.orb:play()
  world.world:killEntity(orb.entity)
  table.insert(world.inventory.orbs, {
    name = "water",
    sprite = G.animRepo:newSprite("WaterOrb_Idle")
  })
  orb_msg(world, "Water")
end

function event_touchCoin(world, player, coin)
  if not coin.entity.collected and not world.playerDead then
    --world.world:killEntity(coin.entity)
    coin.entity:getComponent("Sprite").sprite:play("Empty")
    coin.entity:removeComponent("Highlight")
    world.collectedCoins = world.collectedCoins + 1
    table.insert(world.unsaveCoins, coin)
    coin.entity.collected = true
    G.sounds.coin:play()
  end
end


function event_collectAirOrb(world, player, orb)
  G.sounds.orb:play()
  world.world:killEntity(orb.entity)
  table.insert(world.inventory.orbs, {
    name = "air",
    sprite = G.animRepo:newSprite("AirOrb_Idle")
  })
  orb_msg(world, "Air")
end


local transmuteData = {
  fire = {
    name = "Fire orb",
    message = "You transmuted the fire elemental orb!\nThe potion you made from it burned all the way down your throat but now you can cast fireballs.\nThey could be useful to break old, brittle brick walls! Use [X] on your keyboard of gamepad.\n\n"
  },
  earth = {
    name = "Earth orb",
    message = "You transmuted the earth elemental orb and brewed a dark potion. It tastes like time itself, old and eternal. According to the old legends, you should now be strong enough to perform wall jumps!",
    next_fun = function()
      G.world:show_message("You found the primary elemental orbs! Now we only need the ether orb. It must be hidden somewhere. Maybe if we look in places that we couldn't reach without the Earth potion?")
    end

  },
  water = {
    name = "Water orb",
    message = "You thought the water elemental orb had just turned into water. But as you drink it, you feel the difference. This potion gave you the essence of water. You will never feel thirsty again, but more importantly, fire cannot hurt you anymore!"
  },
  air = {
    name = "Air orb",
    message = "As you transmute the air elemental orb, it seems to vanish. But the air around you smells a lot fresher now and you think it also got a bit colder. After a few minutes you feel much lighter and you notice, that you can peform double jumps now, simply by flapping your arms!"
  },
  ether = {
    name = "Ether orb",
    message = "You found the ether orb and you used it together with the other elemental orbs to produce the philosopher's stone.",
    next_fun = function()
      G.world:show_message("Sadly, as you take it into your hand, you turn into gold.",
        function()
          G.world:show_message("This is the only ending this game has. It was meant for the lulz.", function()
            G.world:show_message("This was Flappy Beard.\nMade by rnlf in 48h\n\n    for Alakajam! #1\n\nThanks for playing!")
          end)

        end)
    end

  }
}


function event_touchLab(world, player, lab)
  if world.transmuting then
    return
  end
  for i, o in ipairs(world.inventory.orbs) do
    if not world.inventory.powers[o.name] then
      --world.inventory.powers[o.name] = true
      world:transmute(transmuteData[o.name].name, function()
        world.inventory.powers[o.name] = true
        world:show_message(transmuteData[o.name].message, transmuteData[o.name].next_fun)
      end)
      break
    end
  end
  if world.activeCP.cp then
    world.activeCP.cp:getComponent("Sprite").sprite:play("CP_Off", false)
    world.activeCP.cp:getComponent("Highlight").color = {128, 128, 128, 255}
  end

  local t = lab.entity:getComponent("Transform")
  world.activeCP.cp = nil
  world.activeCP.pos = {t.pos.x, t.pos.y - 5}
  world.unsaveCoins = {}
end


function event_touchFire(world, player, fire)
  if not world.playerDead and not world.inventory.powers.water then
    kill_player(world, player)
    G.sounds.burn:play()
  end
end

function kill_player2(world, player)

    player:getComponent("Sprite").sprite:play("Player_Die", true):onEnd(function()
      world.playerDead = false
      G.sounds.respawn:play()
      player:getComponent("Transform").pos.x = world.activeCP.pos[1]
      player:getComponent("Transform").pos.y = world.activeCP.pos[2]
      player:getComponent("Transform").last_pos = player:getComponent("Transform").pos:clone()
      for i, c in ipairs(world.unsaveCoins) do
        c.entity:getComponent("Sprite").sprite:play("Coin")
        c.entity:makeComponent("Highlight", 32, 32)
        world.collectedCoins = world.collectedCoins - 1
        c.entity.collected = false
      end
      world.unsaveCoins = {}
    end)
    world.playerDead = true
    local v = player:getComponent("Velocity")
    v.vel.x = 0
    v.acc.x = 0
end


function kill_player(world, player)
  kill_player2(world, player.entity)
end

function event_touchSpike(world, player, fire)
  if not world.playerDead then
    G.sounds.stab:play()
    kill_player(world, player)
  end
end


function event_touchCP(world, player, cp)
  local old = world.activeCP.cp
  if world.activeCP.cp then
    world.activeCP.cp:getComponent("Sprite").sprite:play("CP_Off", false)
    world.activeCP.cp:getComponent("Highlight").color = {128, 128, 128, 255}
  end
  world.activeCP.cp = cp.entity
  world.activeCP.cp:getComponent("Highlight").color = {64, 64, 255, 255}
  cp.entity:getComponent("Sprite").sprite:play("CP_On", true)
  local t = cp.entity:getComponent("Transform")
  world.activeCP.pos = {t.pos.x, t.pos.y-5}
  if old ~= cp.entity or #world.unsaveCoins > 0 then
    G.sounds.checkpoint:play()
  end
  world.unsaveCoins = {}

end
