PhysicsUpdater = coup.class({})


function PhysicsUpdater.updateEntity(e)
  local v = e:getComponent("Velocity")
  local tr = e:getComponent("Transform")
  local po = e:getComponent("PhysicsObject")

  po.collider:moveTo(tr.pos.x + po.x, tr.pos.y + po.y)

  local sens = e:getComponent("Sensors")
  if sens then
    for k, s in pairs(sens.sensors) do
      s.sensor:moveTo(tr.pos.x+s.x, tr.pos.y+s.y)
    end
  end
end


function PhysicsUpdater:update(dt)

  self.world.world:foreachEntityWithComponents(function(e)
    local v = e:getComponent("Velocity")
    local tr = e:getComponent("Transform")
    local po = e:getComponent("PhysicsObject")

    tr.last_pos = tr.pos:clone()

    if not po.dont then
      v.vel = v.vel + v.acc * dt

      if v.acc.x == 0 then
        -- drag
        local drag = vel.hor_drag
        if v.vel.x < 0 then
          v.vel.x = v.vel.x + drag*dt
          if v.vel.x > 0 then
            v.vel.x = 0
          end
        end

        if v.vel.x > 0 then
          v.vel.x = v.vel.x - drag*dt
          if v.vel.x < 0 then
            v.vel.x = 0
          end
        end
      end
    
      if v.vel.x > v.max_vel.x then
        v.vel.x = v.max_vel.x
      elseif v.vel.x < -v.max_vel.x then
        v.vel.x = -v.max_vel.x
      end

      if v.vel.y > v.max_vel.y then
        v.vel.y = v.max_vel.y
      elseif v.vel.y < -v.max_vel.y then
        v.vel.y = -v.max_vel.y
      end

      ---[[
      if po.passivateOnStop then
        if po.passiveCount and math.abs(v.vel.y) < 0.01 and math.abs(v.vel.x) < 0.01 then
          po.passiveCount = po.passiveCount + 1
        else
          po.passiveCount = 0
        end

        if po.passiveCount > 10 then
          G.world.collider:setPassive(po.collider)
          po.dont = true
        end
      end
      --]]

      v.vel.y = v.vel.y + dt * self.gravity * v.grav
      tr.pos = tr.pos + v.vel * dt
    end

    PhysicsUpdater.updateEntity(e)

  end, "Velocity", "Transform", "PhysicsObject")

  self.colC = 0
  self.count = 0
  self.world.collider:update(dt)
  --print(self.colC)
end


local function resolveCollision(s, dx, dy, gndS)
  if dx == 0 and dy == 0 then
    return
  end

  local c = s.owner
  if not c then
    return
  end

  if c.onGroundCollision then
    c.onGroundCollision(gndS)
  end

  local e = c.entity
  local v = e:getComponent("Velocity")
  local tr = e:getComponent("Transform")

  tr.pos.x = tr.pos.x + dx
  tr.pos.y = tr.pos.y + dy

  local l = math.sqrt(dx*dx+dy*dy)
  local nx, ny = dx / l, dy / l

  local n_dot_v = nx*v.vel.x + ny*v.vel.y

  local nnx = nx * n_dot_v
  local nny = ny * n_dot_v

  if (v.vel.y > 30) and (v.vel.y > math.abs(v.vel.x)) then
    G.sounds.land:play()
  end
  v.vel.x = v.vel.x - nnx
  v.vel.y = v.vel.y - nny


  PhysicsUpdater.updateEntity(e)
end


function PhysicsUpdater:new(world)
  self.gravity = 1000
  self.world = world

  self.collisions = {}
  for i = 1, 2000 do
    self.collisions[i] = {}
  end
  self.colC = 0
  self.world.collider:setCallbacks(function(dt, s1, s2, dx, dy)
    self.count = self.count + 1
    if s1.islevel then
      resolveCollision(s2, -dx, -dy, s1)
    elseif s2.islevel then
      resolveCollision(s1, dx, dy, s2)
    else
      if s1.owner and s2.owner then
        self.colC = self.colC + 1
        local cc = self.collisions[self.colC]
        --self.collisions[#self.collisions+1] = {dt, s1.owner, s2.owner, dx, dy}
        cc[1] = dt
        cc[2] = s1.owner
        cc[3] = s2.owner
        cc[4] = dx
        cc[5] = dy
      end
    end
  end)
  --self.collisions = {}
end
