PlayerProcessor = coup.class({})

function PlayerProcessor:new(world)
  self.world = world
end

function PlayerProcessor:update()
  self.world.world:foreachEntityWithComponents(function(e)
    inp = e:getComponent("Input")
    vel = e:getComponent("Velocity")
    sen = e:getComponent("Sensors")
    spr = e:getComponent("Sprite")
    tra = e:getComponent("Transform")

    local onGround = sen.sensors.foot:touchesGround(self.world.collider)
    local onLeft = sen.sensors.left:touchesGround(self.world.collider)
    local onRight = sen.sensors.right:touchesGround(self.world.collider)


    local acc_done = false
    if not onLeft and inp.left then
      vel.acc.x = -2000
      acc_done = true
    end

    if not onRight and inp.right then
      vel.acc.x =  2000
      acc_done = true
    end

    if not acc_done then
      vel.acc.x = 0
    end

    if inp.jump and not inp.lastJump then
      sounds = true
      if onGround then
        vel.vel.y = -300
        inp.allowDouble = true
      elseif onLeft and self.world.inventory.powers.earth then
        vel.vel.y = -300
        vel.vel.x = 300
        inp.allowDouble = true
      elseif onRight and self.world.inventory.powers.earth then
        vel.vel.y = -300
        vel.vel.x = -300
        inp.allowDouble = true
      elseif inp.allowDouble and self.world.inventory.powers.air then
        vel.vel.y = -300
        inp.allowDouble = false
      else 
        sounds = false
      end

      if  sounds then
        G.sounds.jump:play()
      end

    end

    vel.acc.y = 0
    if not onGround then
      vel.hor_drag = 200
      if vel.vel.y < 0 then
        if inp.jump then
          vel.acc.y = -200 -- counteract a bit of gravity
        else
          vel.acc.y = 1500
        end
      end
    else
      vel.hor_drag = 1000
    end

    inp.lastJump = inp.jump


    if onGround  then
      if vel.vel.x > 0 then
        spr.sprite:play("Walk_R", true)
        inp.last_dir = "R"
      elseif vel.vel.x < 0 then
        spr.sprite:play("Walk_L", true)
        inp.last_dir = "L"
      else
        spr.sprite:play("Idle_" .. inp.last_dir, true)
      end
    else
      local updown = "Drop_";
      if vel.vel.y < 0 then
        updown = "Jump_"
      end
      if vel.vel.x > 0 then
        spr.sprite:play(updown .. "R", true)
        inp.last_dir = "R"
      elseif vel.vel.x < 0 then
        spr.sprite:play(updown .. "L", true)
        inp.last_dir = "L"
      else
        spr.sprite:play(updown .. inp.last_dir, true)
      end
    end

    if inp.fire then
      if self.world.inventory.powers.fire then
        if not inp.last_fire or inp.last_fire < self.world.totalTime - 0.5 then
          G.sounds.fireball:play()
          inp.last_fire = self.world.totalTime
          local fb = self.world.world:newEntity()
           
          fb:makeComponent("Transform", tra.pos.x, tra.pos.y - 12)
          local po = fb:makeComponent("PhysicsObject", self.world.collider, 0, 0, 8, 8)
          local vcc
          local sp
          if inp.last_dir == "L" then
            vcc = fb:makeComponent("Velocity", -300, 0)
            sp = fb:makeComponent("Sprite", G.animRepo:newSprite("Fireball_Left"), 1, 8, 8)
          else
            vcc = fb:makeComponent("Velocity", 300, 0)
            sp = fb:makeComponent("Sprite", G.animRepo:newSprite("Fireball_Right"), 1, 8, 8)
          end

          po.onGroundCollision = function(ground)
            po.dont = true
            sp.sprite:play("Fireball_Die"):onEnd(function()
              self.world.world:killEntity(fb)
            end)
            G.sounds.fbdie:play()
            
            G.world.world:foreachEntityWithComponents(function(e)
              local po = e:getComponent("PhysicsObject")
              if po.passivateOnStop and po.dont then
                po.dont = false
                G.world.collider:setActive(po.collider)
              end
            end, "PhysicsObject")

            if ground.owner then
              if ground.owner.item == "breakable" then
                local gndP = ground.owner.entity:getComponent("Transform").pos

                for i = 1, 7 do
                  local deb = self.world.world:newEntity()
                  deb:makeComponent("Transform", math.random(gndP.x - 4, gndP.x + 4), math.random(gndP.y - 12, gndP.y-4))
                  local po = deb:makeComponent("PhysicsObject", self.world.collider, 0, 0, 1, 1)
                  --self.world.collider:addToGroup("player", po.collider)
                  vcc = deb:makeComponent("Velocity", math.random(-5, 20) / 5 * vcc.vel.x, math.random(-200, 200))
                  sp = deb:makeComponent("Sprite", G.animRepo:newSprite("Breakable_Debris"), 1, 24, 8)
                  po.passivateOnStop = true

                end

                self.world.world:killEntity(ground.owner.entity)
              end
            end

          end
          vcc.max_vel.x = 1000
          vcc.hor_drag = 0
          vcc.acc.x = 0.000001 -- workaround for bug
          vcc.grav = 0
          local hl = fb:makeComponent("Highlight", 48, 48, {251,223,155,255})
        end
      end
    end

  end, "Input", "Transform", "Velocity", "Sprite", "Sensors")

end
