local debug = false


local Level = coup.class({})
require "events"
require "physics_updater"
require "tilemap_collider"
require "player_controller"
require "player_processor"
HighlightRenderer = require "highlight_renderer"
require "rain"
--require "sensor_updater"
SpriteRenderer = require "spriterenderer"
local buildCollisionMesh = require "tilemap_collision_builder"

local worldFilename = "data/world.lua"

function loadTiles(name)
  local filename = "data/" .. name
  return G.loadImage(filename)
end

function Level:createObjects(objs)
  local objects = objs.objects.objects
  for i, v in ipairs(objects) do
    
    if v.type == "player" then
      local e = self.world:newEntity("player")
      e:makeComponent("Transform",   v.x, v.y )
      e:makeComponent("Sprite",      G.animRepo:newSprite("Idle_R"), 2, 16, 32)
      e:makeComponent("Velocity", 0, 0)
      --e:makeComponent("PhysicsObject", self.collider, -1, -10, 11.5, 19)
      local po = e:makeComponent("PhysicsObject", self.collider, -1, -10, 
        {
          -9/2, -19/2,
          9/2, -19/2,
          11.5/2, 19/2,
          -11.5/2, 19/2
        }
      )
      po.item = "player"
      e:makeComponent("Input")
      local sensors = e:makeComponent("Sensors")
      self.collider:addToGroup("player", sensors:addSensor(self.collider, "foot", -1, 0, 11, 3))
      self.collider:addToGroup("player", sensors:addSensor(self.collider, "left", -6, -19/2, 3, 17))
      self.collider:addToGroup("player", sensors:addSensor(self.collider, "right", 5, -19/2, 3, 17))
      self.collider:addToGroup("player", po.collider)
      self.activeCP.pos = {v.x, v.y}
      self.player = e

    elseif v.type == "earth_orb" then
      local e = self.world:newEntity("earth_orb")
      e:makeComponent("Transform", v.x, v.y)
      e:makeComponent("Sprite", G.animRepo:newSprite("EarthOrb_Idle"), 1, 8, 8)
      local po = e:makeComponent("PhysicsObject", self.collider, 0, 0, 8, 8)
      local v = e:makeComponent("Velocity", 0, 0)
      self.collider:setPassive(po.collider)
      v.grav = 0
      po.item = "earth_orb"
      po.dont = true
      local hl = e:makeComponent("Highlight", 48, 48)

    elseif v.type == "ether_orb" then
      local e = self.world:newEntity("ether_orb")
      e:makeComponent("Transform", v.x, v.y)
      e:makeComponent("Sprite", G.animRepo:newSprite("EtherOrb_Idle"), 1, 8, 8)
      local po = e:makeComponent("PhysicsObject", self.collider, 0, 0, 8, 8)
      local v = e:makeComponent("Velocity", 0, 0)
      self.collider:setPassive(po.collider)
      v.grav = 0
      po.item = "ether_orb"
      po.dont = true
      local hl = e:makeComponent("Highlight", 48, 48)


    elseif v.type == "fire_orb" then
      local e = self.world:newEntity("fire_orb")
      e:makeComponent("Transform", v.x, v.y)
      e:makeComponent("Sprite", G.animRepo:newSprite("FireOrb_Idle"), 1, 8, 8)
      local po = e:makeComponent("PhysicsObject", self.collider, 0, 0, 8, 8)
      self.collider:setPassive(po.collider)
      local v = e:makeComponent("Velocity", 0, 0)
      v.grav = 0
      po.item = "fire_orb"
      po.dont = true
      local hl = e:makeComponent("Highlight", 48, 48)

    elseif v.type == "water_orb" then
      local e = self.world:newEntity("water_orb")
      e:makeComponent("Transform", v.x, v.y)
      e:makeComponent("Sprite", G.animRepo:newSprite("WaterOrb_Idle"), 1, 8, 8)
      local po = e:makeComponent("PhysicsObject", self.collider, 0, 0, 8, 8)
      local v = e:makeComponent("Velocity", 0, 0)
      self.collider:setPassive(po.collider)
      v.grav = 0
      po.item = "water_orb"
      po.dont = true
      local hl = e:makeComponent("Highlight", 48, 48)

    elseif v.type == "air_orb" then
      local e = self.world:newEntity("air_orb")
      e:makeComponent("Transform", v.x, v.y)
      e:makeComponent("Sprite", G.animRepo:newSprite("AirOrb_Idle"), 1, 8, 8)
      local po = e:makeComponent("PhysicsObject", self.collider, 0, 0, 8, 8)
      self.collider:setPassive(po.collider)
      local v = e:makeComponent("Velocity", 0, 0)
      v.grav = 0
      po.item = "air_orb"
      po.dont = true
      local hl = e:makeComponent("Highlight", 48, 48)

    elseif v.type == "coin" then
      local e = self.world:newEntity()
      e:makeComponent("Transform", v.x, v.y)
      e:makeComponent("Sprite", G.animRepo:newSprite("Coin"), 1, 8, 8)
      local po = e:makeComponent("PhysicsObject", self.collider, 0, 0, 4, 4)
      local v = e:makeComponent("Velocity", 0, 0)
      v.grav = 0
      po.item = "coin"
      self.collider:setPassive(po.collider)
      po.dont = true
      local hl = e:makeComponent("Highlight", 32, 32)
      self.totalCoins = self.totalCoins + 1

    elseif v.type == "lab" then
      local e = self.world:newEntity("lab")
      e:makeComponent("Transform", v.x, v.y)
      e:makeComponent("Sprite", G.animRepo:newSprite("Lab_Idle"), 1, 16, 32)
      local po = e:makeComponent("PhysicsObject", self.collider, 0, -8, 16,16)
      local v = e:makeComponent("Velocity", 0, 0)
      --v.grav = 0
      po.item = "lab"
      po.dont = true
      self.collider:setPassive(po.collider)
      local hl = e:makeComponent("Highlight", 96, 96)

    elseif v.type == "fire" then
      local e = self.world:newEntity()
      e:makeComponent("Transform", v.x, v.y)
      e:makeComponent("Sprite", G.animRepo:newSprite("Fire_Burn"), 3, 8, 16)
      local po = e:makeComponent("PhysicsObject", self.collider, 0, -4, 12,8)
      local v = e:makeComponent("Velocity", 0, 0)
      --v.grav = 0
      po.item = "fire"
      po.dont = true
      self.collider:setPassive(po.collider)
      local hl = e:makeComponent("Highlight", 96, 96, {251,223,155,255})

    elseif v.type == "spikes" then
      local e = self.world:newEntity()
      e:makeComponent("Transform", v.x, v.y)
      local spike_data = {
        up = {
          anim = "Up",
          spr_off = {8, 16},
          hb = {0, -4, 12, 8}
        },
        left = {
          anim = "L",
          spr_off = {16, 8},
          hb = {-4, 0, 8, 12}
        },
        right = {
          anim = "R",
          spr_off = {0, 8},
          hb = {4, 0, 8, 12}
        },
        down = {
          anim = "Down",
          spr_off = {8, 0},
          hb = {0, 4, 12, 8}
        }

      }
      local dir = v.properties.dir
      local spd = spike_data[dir]
      local n = math.random(0,1)
      e:makeComponent("Sprite", G.animRepo:newSprite("Spikes_" .. spd.anim .. "_" .. tostring(n)), 3, unpack(spd.spr_off))
      local po = e:makeComponent("PhysicsObject", self.collider, unpack(spd.hb))
      local v = e:makeComponent("Velocity", 0, 0)
      po.item = "spike"
      po.dont = true
      self.collider:setPassive(po.collider)

    elseif v.type == "cp" then
      local e = self.world:newEntity()
      e:makeComponent("Transform", v.x, v.y)
      e:makeComponent("Sprite", G.animRepo:newSprite("CP_Off"), 0, 8, 16)
      local po = e:makeComponent("PhysicsObject", self.collider, 0, -8, 10,16)
      local v = e:makeComponent("Velocity", 0, 0)
      --v.grav = 0
      po.item = "cp"
      po.dont = true
      self.collider:setPassive(po.collider)
      local hl = e:makeComponent("Highlight", 96, 96, {128,128,128,255})

    elseif v.type == "breakable" then
      local e = self.world:newEntity()
      e:makeComponent("Transform", v.x, v.y)
      e:makeComponent("Sprite", G.animRepo:newSprite("Breakable_Exist"), 3, 24, 16)
      local po = e:makeComponent("PhysicsObject", self.collider, 0, -8, 16,16)
      self.collider:addToGroup("level", po.collider)
      po.collider.islevel = true
      local v = e:makeComponent("Velocity", 0, 0)
      --v.grav = 0
      po.item = "breakable"
      po.dont = true
      self.collider:setPassive(po.collider)

    end
  end
end



function Level:new()
  self.unsaveCoins = {}
  self.totalCoins = 0
  self.totalTime = 0
  self.collectedCoins = 0
  local objs, imgs, props
  self.t = 0

  self.tilemap, objs, imgs, props = coup.tiledloader.loadTiledMapFile(worldFilename, loadTiles)

  self.collider = HC.new(32)
  self.activeCP = {}
  
  self.mainLayer = self.tilemap:get_layer("main")

  buildCollisionMesh(self.collider, self.tilemap.width, self.tilemap.height, self.tilemap.tilewidth, self.tilemap.tileheight,
    function(x, y)
      return self.mainLayer:get_tile(x-1, y-1)
    end,
    function(a)
      return a == 0
    end)

  self.mainLayerRenderer = coup.tilelayerrenderer.TilelayerRenderer(self.mainLayer, 10)

  self.bgLayer = self.tilemap:get_layer("background")
  self.bgLayerRenderer = coup.tilelayerrenderer.TilelayerRenderer(self.bgLayer, 10)
  self.overlayLayer = self.tilemap:get_layer("overlay")
  self.overlayLayerRenderer = coup.tilelayerrenderer.TilelayerRenderer(self.overlayLayer, 10)

  self.tween = coup.tween.group()
  self.world = coup.world.World()
  coup.components.register(self.world)
  self.world:registerComponentType("PhysicsObject", require("physics_object")())
  self.world:registerComponentType("Input", require("input")())
  self.world:registerComponentType("Sensors", require("sensors")())
  self.world:registerComponentType("Highlight", require("highlight")())
  self:createObjects(objs)

  self.camera = self.world:newEntity("camera")
  local w, h = G.screenSize[1], G.screenSize[2]
  self.camera:makeComponent("Transform", w/2, h/2)
  self.camera:makeComponent("BoundingBox", -w/2, -h/2, w/2, h/2)

  --self.spriteSpatialHash     = coup.spatialhash.SpatialHashtable(50)
  --self.spriteAutoHasher      = coup.systems.AutoHasher(self.spriteSpatialHash, self.world, "Sprite")
  self.spriteRenderer        = SpriteRenderer(self.collider)
  self.spriteUpdater         = coup.systems.SpriteUpdater(self.world)
  self.physicsUpdater        = PhysicsUpdater(self)
  self.highlightRenderer  = HighlightRenderer(self.collider)
--  self.sensorUpdater         = SensorUpdater(self)

  self.playerController = PlayerController(self.player)
  self.playerProcessor = PlayerProcessor(self)

  self.rain = Rain()
  
  self.world:step()

  self.vignette = G.loadImage("data/vignette.png")
  self.lightCanvas = love.graphics.newCanvas(G.screenSize[1], G.screenSize[2])

  self.inventory = {
    orbs   = {},
    powers = {}
  }

  self.timerTween = {t = 0}
  self.tween:tween(self.timerTween, 0.5, {t = 1}):on_complete(function()
    self:show_message("Welcome, friend!\nMy name is Flappy Beard and I am an alchemist!\nMy plan is to create the philosopher's stone today. To do that, I will have to find all five elemental orbs.", function()
      self:show_message("Each of the orbs will also allow me to brew a potion that will give me new abilities. Those should come in handy to find the remaining ones!", function()
        self:show_message("Walk around with [LEFT] and [RIGHT] or your [DPAD]. Jump with [C] on your keyboard or [A] on your gamepad.")
      end)
    end)
  end)

  self.coinSprite = G.animRepo:newSprite("Coin")

end


function Level:updateCamera(lerp)
  local p = self.camera:getComponent("Transform")
  local pp = self.player:getComponent("Transform")

  if not debug and not self.transmuting and not self.playerDead then
    p.pos = pp:lpos(lerp):clone()
  end
  AABBUpdater_updateEntity(self.camera)
  --print(p.vel.x, p.acc.x, p.hor_drag)

end


local collisionEvents = {
  {"player", "fire_orb", event_collectFireOrb},
  {"player", "earth_orb", event_collectEarthOrb},
  {"player", "water_orb", event_collectWaterOrb},
  {"player", "air_orb", event_collectAirOrb},
  {"player", "ether_orb", event_collectEtherOrb},
  {"player", "lab", event_touchLab},
  {"player", "fire", event_touchFire},
  {"player", "spike", event_touchSpike},
  {"player", "coin", event_touchCoin},
  {"player", "cp", event_touchCP}
}


function Level:handleCollision(a, b)
  for i, c in ipairs(collisionEvents) do
    if a.item == c[1] and b.item == c[2] then
      c[3](self, a, b)
    elseif a.item == c[2] and b.item == c[1] then
      c[3](self, b, a)
    end
  end
end


local timestep = 1 / 120
function Level:update(dt)
  if not self.message then
    self.t = self.t + dt

    local linearStep = 0
    while self.t > timestep do
      if not self.transmuting then
        self.playerController:update()
        if not self.playerDead then
          self.playerProcessor:update()
        end
        self.physicsUpdater:update(timestep)
      end
      --for i, c in ipairs(self.physicsUpdater.collisions) do
      for i = 1, self.physicsUpdater.colC do
        local c = self.physicsUpdater.collisions[i]
        self:handleCollision(c[2], c[3])
      end
      local y = self.player:getComponent("Transform").pos.y
      if y > 140 * 16 then
        if not self.playerDead then
          G.sounds.fall:play()
        end
        kill_player2(self, self.player)
      end
      linearStep = linearStep + timestep
      self.t = self.t - timestep
      self.world:step()
    end


    self.totalTime = self.totalTime + linearStep
    self:updateCamera(self.t / timestep)
    self.rain:update(linearStep)
    self.spriteUpdater:updateSprites(linearStep)
    for k, s in pairs(self.inventory.orbs) do
      s.sprite:update(linearStep)
    end
    self.tween:update(linearStep)
  else
    self.tween:update(dt)
    if self.message then
      if self.message.size == 1 then
        local gpClose = love.keyboard.isDown("return")
        local joysticks = love.joystick.getJoysticks()
        for i, joystick in ipairs(joysticks) do
          gpClose = gpClose or joystick:isGamepadDown('b')
        end
        if gpClose then
          self.tween:tween(self.message, 0.2, {size = 0}):on_complete(function()
            local f = self.message.next_func
            self.message = nil
            if f then
              f()
            end
          end)
    
        end
      end
    end
  end

end


function Level:show_message(message, next_func)
  ---[[
  self.message = {
    text = message,
    size = 0,
    next_func = next_func
  }

  G.sounds.blip:play()

  self.tween:tween(self.message, 0.3, {size = 1})
  --]]
end


function Level:debug_transform()
  self.world:foreachEntityWithComponents(function(e)
    local t = e:getComponent("Transform")
    love.graphics.line(t.pos.x - 4, t.pos.y - 4, t.pos.x + 4, t.pos.y + 4)
    love.graphics.line(t.pos.x + 4, t.pos.y - 4, t.pos.x - 4, t.pos.y + 4)
  end, "Transform")
end

function Level:debug_entity_names()
  self.world:foreachEntityWithComponents(function(e)
    if e.name then
      local t = e:getComponent("Transform")
      love.graphics.print(e.name, t.pos.x + 5, t.pos.y)
    end
  end, "Transform")
end

function Level:debug_hc()
  love.graphics.setLineStyle('smooth')
  for k, v in pairs(self.collider:shapesInRange(0,0, self.tilemap.width, self.tilemap.height)) do
    v:draw()
  end
end

function Level:draw()

  love.graphics.clear(30, 20, 50, 255)

  local lerp = self.t / timestep
  local cab = self.camera:getComponent("AxisAlignedBoundingBox")
  self.bgLayerRenderer:draw(cab.topLeft, cab.bottomRight)

  love.graphics.setBlendMode('add')
  self.highlightRenderer:renderVisible(self.camera, lerp, {64,64,64,255}, true)
  love.graphics.setBlendMode('alpha')
  love.graphics.setColor(255,255,255,255)

  self.mainLayerRenderer:draw(cab.topLeft, cab.bottomRight)
  self.spriteRenderer:renderVisible(self.camera, lerp)

  self.overlayLayerRenderer:draw(cab.topLeft, cab.bottomRight)

  self.rain:render()
  local can = love.graphics.getCanvas()
  love.graphics.setCanvas(self.lightCanvas)
  love.graphics.clear(0,0,0,255)


  love.graphics.setBlendMode('add')
  love.graphics.setColor(255,255,255,255)
  love.graphics.draw(self.vignette)
  self.highlightRenderer:renderVisible(self.camera, lerp)

  love.graphics.setCanvas(can)
  love.graphics.setBlendMode('multiply')
  love.graphics.draw(self.lightCanvas)

  love.graphics.setBlendMode('alpha')

  self:drawInventory()

  self:drawMessage()
  self:drawTransmuting()

  if debug then
    self:debug_transform()
    self:debug_entity_names()
    self:debug_hc()
  end

end


function Level:drawMessage()
  if self.message then
    local w = math.floor(G.screenSize[1] * 0.8 * self.message.size)
    local h = math.floor(G.screenSize[2] * 0.4 * self.message.size)
    love.graphics.setColor(0,0,0,224)
    love.graphics.setLineWidth(2)
    love.graphics.rectangle('fill', (G.screenSize[1] - w) / 2, G.screenSize[2] * 0.75 - h / 2, w, h)
    love.graphics.setColor(255,255,255,255)
    love.graphics.rectangle('line', (G.screenSize[1] - w) / 2, G.screenSize[2] * 0.75 - h / 2, w, h)

    if self.message.size == 1 then
      love.graphics.printf(self.message.text,
        (G.screenSize[1] - w) / 2 + 5,
         G.screenSize[2] * 0.75 - h / 2 + 5,
         w-10,
         "left")

      love.graphics.setColor(128,128,128,255)
      love.graphics.printf("Hit [RETURN] or [B] on Gamepad", 
        (G.screenSize[1] - w) / 2 + 5,
         G.screenSize[2] * 0.75 + h / 2 - 10,
         w-10,
         "right")
      love.graphics.setColor(255,255,255,255)
    end

  end
end


function Level:drawTransmuting()
  if self.transmuting then
    love.graphics.setColor(0,0,0,224)
    love.graphics.rectangle("fill", 50, G.screenSize[2] - 50, G.screenSize[1] - 100, 20)
    love.graphics.setColor(255,255,255,255)
    love.graphics.setLineWidth(2)
    love.graphics.rectangle("line", 50, G.screenSize[2] - 50, G.screenSize[1] - 100, 20)
    love.graphics.rectangle("fill", 50, G.screenSize[2] - 50,
        (G.screenSize[1] - 100 ) * self.transmuting.time,
        20)
    love.graphics.printf("Transmuting " .. self.transmuting.name, 50, G.screenSize[2] - 45, G.screenSize[1] - 100, "center")
    love.graphics.setColor(255,255,255,255)
  end
end



function Level:transmute(name, func)
  self.transmuting = {
    time = 0,
    name = name,
    func = func
  }

  local pt = self.player:getComponent("Transform")
  local lt = self.world:getEntityByName("lab"):getComponent("Transform")
  pt.pos = lt.pos:clone()
  pt.last_pos = lt.pos:clone()

  local pv = self.player:getComponent("Velocity")
  pv.vel.x = 0
  pv.vel.y = 0

  local ps = self.player:getComponent("Sprite")
  ps.sprite:play("Drink_Potion")

  self.tween:tween(self.transmuting, 3, {time = 1}):on_complete(function()
    local f = self.transmuting.func
    self.transmuting = nil

    if f then f() end
  end)

  

end


function Level:drawInventory()
  local unmuated = false
  for i, s in ipairs(self.inventory.orbs) do
    local f = s.sprite:getCurrentFrame()
    if self.inventory.powers[s.name] then
      love.graphics.setColor(255,255,255,255)
    else
      love.graphics.setColor(255,64,64,128)
      unmuated = true
    end
    love.graphics.draw(f.img, f.quad, (i-1)*16,0, 0, 1, 1, f.origin.x, f.origin.y)
  end
  love.graphics.setColor(255,255,255,255)

  local cf = self.coinSprite:getCurrentFrame()
  love.graphics.draw(cf.img, cf.quad, G.screenSize[1] - 16,0, 0, 1, 1, cf.origin.x, cf.origin.y)
  love.graphics.printf(self.collectedCoins .. "/" .. self.totalCoins, 0, 4, G.screenSize[1] - 16, "right")




end


return Level
